"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.appendDeclareSvelteVarsTypes = void 0;
const traverse_1 = require("../../traverse");
const script_1 = require("../script");
/**
 * Append type declarations for svelte variables.
 * - Append TypeScript code like
 *   `declare let $foo: Parameters<Parameters<(typeof foo)["subscribe"]>[0]>[0];`
 *   to define the type information for like `$foo` variable.
 * - Append TypeScript code like `let foo = bar;` to define the type information for like `$: foo = bar` variable.
 */
function appendDeclareSvelteVarsTypes(ctx) {
    const vcode = ctx.sourceCode.scripts.vcode;
    if (/\$\s*:\s*[\p{ID_Start}$(_]/u.test(vcode)) {
        // Probably have a reactive variable, so we will need to parse TypeScript once to extract the reactive variables.
        const result = (0, script_1.parseScriptWithoutAnalyzeScope)(vcode, ctx.sourceCode.scripts.attrs, Object.assign(Object.assign({}, ctx.parserOptions), { 
            // Without typings
            project: null }));
        appendDeclareSvelteVarsTypesFromAST(result, vcode, ctx);
    }
    else {
        appendDeclareStoreTypesFromText(vcode, ctx);
    }
}
exports.appendDeclareSvelteVarsTypes = appendDeclareSvelteVarsTypes;
/**
 * Append type declarations for svelte variables from AST.
 */
function appendDeclareSvelteVarsTypesFromAST(result, code, ctx) {
    const maybeStores = new Set();
    (0, traverse_1.traverseNodes)(result.ast, {
        visitorKeys: result.visitorKeys,
        enterNode: (node, parent) => {
            if (node.type === "Identifier") {
                if (!node.name.startsWith("$") || node.name.length <= 1) {
                    return;
                }
                maybeStores.add(node.name.slice(1));
            }
            else if (node.type === "LabeledStatement") {
                if (node.label.name !== "$" ||
                    parent !== result.ast ||
                    node.body.type !== "ExpressionStatement" ||
                    node.body.expression.type !== "AssignmentExpression" ||
                    // Must be a pattern that can be used in the LHS of variable declarations.
                    // https://github.com/ota-meshi/svelte-eslint-parser/issues/213
                    (node.body.expression.left.type !== "Identifier" &&
                        node.body.expression.left.type !== "ArrayPattern" &&
                        node.body.expression.left.type !== "ObjectPattern" &&
                        node.body.expression.left.type !== "AssignmentPattern" &&
                        node.body.expression.left.type !== "RestElement")) {
                    return;
                }
                // It is reactive variable declaration.
                const text = code.slice(...node.body.expression.range);
                ctx.scriptLet.appendDeclareReactiveVar(text);
            }
        },
        leaveNode() {
            /* noop */
        },
    });
    ctx.scriptLet.appendDeclareMaybeStores(maybeStores);
}
/**
 * Append type declarations for store access.
 * Append TypeScript code like
 * `declare let $foo: Parameters<Parameters<(typeof foo)["subscribe"]>[0]>[0];`
 * to define the type information for like `$foo` variable.
 */
function appendDeclareStoreTypesFromText(vcode, ctx) {
    const extractStoreRe = /\$[\p{ID_Start}$_][\p{ID_Continue}$\u200c\u200d]*/giu;
    let m;
    const maybeStores = new Set();
    while ((m = extractStoreRe.exec(vcode))) {
        const storeName = m[0];
        const originalName = storeName.slice(1);
        maybeStores.add(originalName);
    }
    ctx.scriptLet.appendDeclareMaybeStores(maybeStores);
}
